import numpy as np
import sys
import xml.etree.ElementTree as ET
import time as clock_time
from os.path import exists, realpath


def get_command_line_arguments():
    """
    Returns a list of files that have been provided as a command line argument
    :return: list of files
    """

    # Parse and assert command line options
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input-data", default=None,
                        help="File with filtered info on cores ID, function ID, and timings")
    parser.add_argument("--ts-number", default=None, type=int,
                        help="Number of time-steps present in the trace")
    parser.add_argument("-t", "--tool-path", default=None,
                        help="Path to Magicut templates")
    parser.add_argument("--function-file", default=None,
                        help="File with names corresponding to function IDs")

    args = parser.parse_args()

    if args.input_data is None:
        parser.error("Input data are needed")
    else:
        if not exists(args.input_data):
            parser.error("The specified input does not exists")
    if args.function_file is not None and not exists(args.function_file):
        parser.error("The specified file  does not exists")
    if args.ts_number is None:
        print("Warning: magicCut will try to guess the number of time-steps")
    if args.tool_path is None:
        args.tool_path = "/".join(realpath(__file__).split("/")[:-2]) + "/templates/"
    if not exists(args.tool_path):
        parser.error("The specified path to tools does not exist: " + args.tool_path)

    return args


def save_function_names(function_ids, functions_name_file):
    # Opens file with format #ID FUNCTION_NAME
    with open(functions_name_file) as f:
        ids_name_list = f.read()
    id = [l.split()[0] for l in ids_name_list.splitlines()]
    names = [l.split()[1] for l in ids_name_list.splitlines()]
    with open('functions.txt', 'w') as f:
        for ID in function_ids:
            index = id.index(str(ID))
            f.write(id[index] + " " + names[index + "\n"])
    print("Functions that are called once per time-step saved to functions.txt")


def cut_time(tool_path, input_data, time_steps=None):
    # Path to the cutter template
    template_path = tool_path + "cutter_template.xml"
    print("Starting the magic")
    start = clock_time.time()

    # Read trace and load function calls
    cpu_ids, time, function_ids = np.genfromtxt(input_data, dtype='int', unpack='True')

    print("for opening the input_data: ")
    end = clock_time.time()
    print(end - start)

    # Find set of different routines, and number of cores
    unique_function_ids, call_counts = np.unique(function_ids, return_counts=True)
    cpu_number = len(np.unique(cpu_ids))

    if time_steps is None:
        # Remember that everything is multiplied by N cores
        # Creates an histogram with the number of many functions are called n times
        counts = np.bincount(call_counts)
        # Suppose that the most common value is the number of ts
        time_steps = np.argmax(counts)
        # Find functions that are called once per time-step
        functions_called_once_per_step = unique_function_ids[call_counts == time_steps]
    else:
        # We just search for that functions that appears once per core per time-step
        functions_called_once_per_step = unique_function_ids[call_counts / cpu_number == time_steps]

    if not len(functions_called_once_per_step):
        raise Exception("No function has been found which is called once per timestep")

    # Find which routine is called in first place
    first_routine = functions_called_once_per_step[0]
    print("First called routine: ", first_routine)

    # Find number of nemo_proc_number
    nemo_proc = cpu_ids[function_ids == first_routine]
    nemo_proc = np.unique(nemo_proc)
    nemo_proc_number = nemo_proc.shape[0]

    nemo_proc_min = min(nemo_proc)
    nemo_proc_max = max(nemo_proc)

    time_steps = time[(function_ids == first_routine) & (cpu_ids == nemo_proc[0])].shape[0]
    print("Actual n_proc", nemo_proc_number, "with ", time_steps, "time steps")

    # Find the index of the fastest step
    time_step = np.ones([nemo_proc_number, time_steps])

    ts_min_index = np.zeros(nemo_proc_number, dtype='int')
    ts_max_index = np.zeros(nemo_proc_number, dtype='int')

    ts_time = np.zeros([nemo_proc_number, time_steps])
    start = clock_time.time()
    for proc in nemo_proc:
        index = proc - nemo_proc[0]
        ts_time[index] = time[(cpu_ids == proc) & (function_ids == first_routine)]

        # compute the duration of each ts
        ts_duration = np.diff(ts_time[index])

        # find the index of the minimum ts for each proc
        ts_min_index[index] = np.argmin(ts_duration)

        # find the index of the maximum ts for each proc
        ts_max_index[index] = np.argmax(ts_duration[1:-1]) + 1

    # Evaluate the most common index for best ts
    counts = np.bincount(ts_min_index)
    best_ts_index = np.argmax(counts)

    counts = np.bincount(ts_max_index)
    worst_ts_index = np.argmax(counts)

    print("for finding the index of the slowest / fastest step: ", worst_ts_index, "/", best_ts_index)
    end = clock_time.time()
    print(end - start)

    # Find the start and the end of the best step
    best_ts_start = min(ts_time[:, best_ts_index])
    best_ts_end = max(ts_time[:, best_ts_index + 1])

    # Find the start and the end of the best step
    worst_ts_start = min(ts_time[:, worst_ts_index])
    worst_ts_end = max(ts_time[:, worst_ts_index + 1])

    print("Worst / best time step's duration: ", worst_ts_end - worst_ts_start, best_ts_end - best_ts_start)

    # Open the xml template
    tree = ET.parse(template_path)

    cutter = tree.find('cutter')
    for tasks in tree.iter('tasks'):
        tasks.text = str(nemo_proc_min) + "-" + str(nemo_proc_max)
    for minimum_time in tree.iter('minimum_time'):
        minimum_time.text = str(best_ts_start - 1000)
    for maximum_time in tree.iter('maximum_time'):
        maximum_time.text = str(best_ts_end + 1000)

    # Create paramedir cutter file
    tree.write('best_time_cutter.xml')
    for minimum_time in tree.iter('minimum_time'):
        minimum_time.text = str(worst_ts_start - 1000)
    for maximum_time in tree.iter('maximum_time'):
        maximum_time.text = str(worst_ts_end + 1000)

    # Create paramedir cutter file
    tree.write('worst_time_cutter.xml')

    return functions_called_once_per_step


if __name__ == "__main__":
    # Get files from command line
    args = get_command_line_arguments()

    tool_path = args.tool_path
    input_data = args.input_data
    function_file = args.function_file
    ts_number = args.ts_number

    unique_functions = cut_time(tool_path, input_data, ts_number)
    save_function_names(unique_functions, function_file)
